# -*- coding: utf-8 -*-

import unittest
from .. import unis3
from os.path import realpath, dirname
from django.core.files.base import ContentFile
from tempfile import NamedTemporaryFile
from django.core.files import File
from ..files import TmpFile


class UniS3TestCase(unittest.TestCase):
    """
    " Unit tests for the unis3 wrapper.
    """

    def setUp(self):
        self.storage = unis3.S3Storage('unittest')
        # file for testing : this file
        self.tst_path = realpath(__file__)
        self.src_path = realpath(unis3.__file__)
        self.ext_path = realpath(unittest.__file__)
        self.current_dir = dirname(self.tst_path)
        self.parent_dir = dirname(self.current_dir)
        self.src_url = self.storage.upload(self.src_path)
        self.tst_url = self.storage.upload(self.tst_path)
        self.ext_url = self.storage.upload(self.ext_path)

    def tearDown(self):
        for f in self.storage.listdir()[1]:
            self.storage.delete(f)
        self.storage.delete_bucket()

    def test_bucket_name(self):
        self.assertEqual('unittest', self.storage.bucket)

    def test_bucket_location(self):
        expected = 'https://s3.unistra.fr/unittest'
        self.assertEqual(expected, self.storage.location)

    def test_exists(self):
        self.assertTrue(self.storage.exists(self.tst_path))
        self.assertTrue(self.storage.exists(self.src_path))
        self.assertTrue(self.storage.exists(self.ext_path))

    def test_urls(self):
        self.assertEqual(self.tst_url, self.storage.url(self.tst_url))
        self.assertEqual(self.src_url, self.storage.url(self.src_url))
        self.assertEqual(self.ext_url, self.storage.url(self.ext_url))

    def test_list_dir_from_root(self):
        directories, files = self.storage.listdir()
        self.assertEqual(len(directories), 0)
        self.assertEqual(len(files), 3)
        self.assertTrue(self.tst_path in files)
        self.assertTrue(self.src_path in files)
        self.assertTrue(self.ext_path in files)

    def test_list_dir_from_current_dir(self):
        directories, files = self.storage.listdir(self.current_dir)
        self.assertEqual(len(directories), 0)
        self.assertEqual(len(files), 1)
        self.assertTrue(self.tst_path in files)
        self.assertFalse(self.src_path in files)
        self.assertFalse(self.ext_path in files)

    def test_list_dir_from_parent_dir(self):
        directories, files = self.storage.listdir(self.parent_dir)
        self.assertEqual(len(directories), 0)
        self.assertEqual(len(files), 2)
        self.assertTrue(self.tst_path in files)
        self.assertTrue(self.src_path in files)
        self.assertFalse(self.ext_path in files)

    def test_reupload(self):
        self.assertEqual(len(self.storage.listdir()[1]), 3)
        self.src_url = self.storage.upload(self.src_path)
        self.tst_url = self.storage.upload(self.tst_path)
        self.ext_url = self.storage.upload(self.ext_path)
        self.assertEqual(len(self.storage.listdir()[1]), 3)

    def test_delete(self):
        self.assertTrue(self.storage.exists(self.tst_path))
        self.assertTrue(self.storage.exists(self.src_path))

        res = self.storage.delete(self.src_path)
        self.assertTrue(res)
        self.assertTrue(self.storage.exists(self.tst_path))
        self.assertFalse(self.storage.exists(self.src_path))
        self.assertEqual(len(self.storage.listdir()[1]), 2) # tst and ext remain

        res = self.storage.delete(self.tst_path)
        self.assertTrue(res)
        self.assertFalse(self.storage.exists(self.tst_path))
        self.assertFalse(self.storage.exists(self.src_path))
        self.assertEqual(len(self.storage.listdir()[1]), 1) # only ext remains

    def test_delete_nonexistent(self):
        path = 'non/existent/path'
        self.assertFalse(self.storage.exists(path))
        directories, files = self.storage.listdir(self.current_dir) # before
        res = self.storage.delete(path)
        self.assertTrue(res) # deleting unexistent stuff always successes
        d, f = self.storage.listdir(self.current_dir) # after
        self.assertEqual(len(d), len(directories))
        self.assertEqual(len(f), len(files))

    def test_available_name_overwrite_mode(self):
        self.storage.overwrite = True
        name = self.storage.get_available_name(self.src_path)
        self.assertEqual(self.src_path, name)

    def test_available_name_no_overwrite(self):
        self.storage.overwrite = False
        name = self.storage.get_available_name(self.src_path)
        self.assertGreater(len(name), len(self.src_path))
        common_part = self.src_path[:-len('.py')]
        self.assertTrue(name.startswith(common_part))

    def test_open(self):
        f = self.storage.open(self.src_url)
        self.assertGreater(len(f.read()), 0)
        f.close()

    def test_open_write_mode_failure(self):
        with self.assertRaises(AttributeError) as context:
            self.storage.open(self.src_url, 'w')
        self.assertTrue("Only read operations" in str(context.exception))
        with self.assertRaises(AttributeError) as context:
            self.storage.open(self.src_url, 'r+')
        self.assertTrue("Only read operations" in str(context.exception))
        with self.assertRaises(AttributeError) as context:
            self.storage.open(self.src_url, 'w+')
        self.assertTrue("Only read operations" in str(context.exception))

    def test_read(self):
        with open(self.src_path) as f:
            before = f.read()
        with unis3.S3FileOpener(self.src_url, 'r') as f:
            after = f.read()
        self.assertEqual(after, before)

    def test_overwrite(self):
        before = 'whatever'
        fd, path = TmpFile.create()
        TmpFile.write(fd, before)
        file = File(open(path, 'rb'))
        url = self.storage.save(path, file)
        file.close()

        with unis3.S3FileOpener(url, 'r') as f:
            data = f.read()
            self.assertEqual(data, before)

        self.assertTrue(self.storage.overwrite)

        after = 'tu peux pas test'
        fd2, path2 = TmpFile.create()
        TmpFile.write(fd2, after)
        f2 = File(open(path2, 'rb'))
        url2 = self.storage.save(path, f2) # path, NOT path2 (overwrite)
        f2.close()
        end_of_common_prefix = 'Signature='
        index = url.find(end_of_common_prefix)+len(end_of_common_prefix)
        self.assertEqual(url2[:index], url[:index])
        with unis3.S3FileOpener(url, 'r') as f:
            data = f.read()
            self.assertEqual(data, after)
        self.assertNotEqual(after, before)



# -*- coding: utf-8 -*-

from tempfile import mkdtemp, mkstemp
from os import fdopen


class TmpFile:

    @staticmethod
    def create(dir=None, filename=None, mode='r'):
        binary = 'b' in mode
        dir = dir or mkdtemp()
        fd, path = filename or mkstemp(dir=dir, text=(not binary))
        return fd, path

    @staticmethod
    def write(fd, data, mode='w'):
        f = fdopen(fd, mode)
        f.write(data)
        f.close()



# -*- coding: utf-8 -*-

from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.functional import cached_property
from s3_local_endpoint import s3_local_endpoint as s3

from urllib.parse import urlparse
import requests
import os

from .files import TmpFile
from django.core.files import File as DjangoNIHFile

AWS_ACCESS_KEY_ID = os.environ['AWS_ACCESS_KEY_ID']
AWS_SECRET_ACCESS_KEY = os.environ['AWS_SECRET_ACCESS_KEY']
AWS_S3_ENDPOINT_URL = os.environ['AWS_S3_ENDPOINT_URL']
AWS_STORAGE_BUCKET_NAME = os.environ['AWS_STORAGE_BUCKET_NAME']


class S3FileOpener:

    class File(DjangoNIHFile):
        def close(self):
            super(S3FileOpener.File, self).close()
            os.remove(self.file.name)


    def __init__(self, url, mode=None):
        self.mode = mode or 'r'
        self.request = requests.get(url, allow_redirects=True)
        fd, path = TmpFile.create()
        TmpFile.write(fd, self.request.content, 'wb')
        self.file = self.File(open(path, self.mode))

    def __enter__(self):
        return self.file

    def __exit__(self, exception_type, exception_value, traceback):
        self.file.close()


@deconstructible
class S3Storage(Storage):

    def __init__(self, bucket=AWS_STORAGE_BUCKET_NAME, overwrite=True):
        """
        " :attr bucket Path to « folder » containing files.
        """
        self.bucket = bucket
        self.client = s3.s3_local_endpoint(
            AWS_ACCESS_KEY_ID,
            AWS_SECRET_ACCESS_KEY,
            AWS_S3_ENDPOINT_URL,
            self.bucket)
        self.overwrite = overwrite

    @cached_property
    def location(self):
        return os.path.join(AWS_S3_ENDPOINT_URL, self.bucket)

    def delete_bucket(self):
        self.client.delete_bucket(self.bucket)

    def delete(self, path):
        assert path, "Path cannot be empty."
        return self.client.delete_file(path)

    def exists(self, path):
        files = self.client.list_files()
        return path in files

    def listdir(self, path=None):
        path = path or ''
        directories, files = [], []
        for f in self.client.list_files():
            if f.startswith(path):
                # list_files doesn't return dirs independantly, ever
                # so we don't have to do something like this, ever:
                # if os.path.isdir(f):
                #   directories.append(f)
                # else:
                files.append(f)
        return directories, files

    def get_available_name(self, name, max_length=None):
        if self.overwrite:
            return name
        return super(S3Storage, self).get_available_name(name, max_length)

    def url(self, name):
        return name

    def _save(self, path, content):
        content.seek(0) # make sure content can be read during upload
        self.client.upload_stream(path, content)
        return self.client.generate_url(path)

    def _open(self, name, mode='rb'):
        if ('w' in mode) or ('+' in mode):
            # writing COULD be done by downloading the file into a TmpFile,
            # then writing in it, then reuploading it at the same url
            # question is : SHOULD it be done ?
            raise AttributeError("Only read operations supported for now.")
        opener = S3FileOpener(name)
        return opener.file

    def upload(self, path):
        f = open(path, 'rb')
        url = self.save(path, f)
        f.close()
        return url



# -*- coding: utf-8 -*-

import os
from setuptools import setup
from setuptools import find_packages


with open('README.rst') as readme:
    long_description = readme.read()


def recursive_requirements(requirement_file, libs, links, path=''):
    if not requirement_file.startswith(path):
        requirement_file = os.path.join(path, requirement_file)
    with open(requirement_file) as requirements:
        for requirement in requirements.readlines():
            if requirement.startswith('-r'):
                requirement_file = requirement.split()[1]
                if not path:
                    path = requirement_file.rsplit('/', 1)[0]
                recursive_requirements(requirement_file, libs, links,
                                       path=path)
            elif requirement.startswith('-f'):
                links.append(requirement.split()[1])
            else:
                libs.append(requirement)

libraries, dependency_links = [], []
recursive_requirements('requirements.txt', libraries, dependency_links)

setup(
    name='django-unis3',
    version='1.0.0',
    packages=find_packages(),
    install_requires=libraries,
    dependency_links=dependency_links,
    long_description=long_description,
    description='S3 Django Storage.',
    author='di-dip-unistra',
    author_email='dnum-dip@unistra.fr',
    maintainer='di-dip-unistra',
    maintainer_email='dnum-dip@unistra.fr',
    url='https://git.unistra.fr/di/cesar/django-unis3.git',
    download_url='',
    license='GPLv3',
    keywords=['django', 's3', 'Université de Strasbourg'],
    include_package_data=True,
)
